/* Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
`include "rtavr_common.v"

`ifdef IOR_HAVE_CPI

// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

// Module Name: rtavr_ior_usart
//   version 005
//
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
module rtavr_ior_cpi # (
		parameter BASE_ADDR = 6'h09  // 0x09: CPCSR , 0x0a : CPDR
) (
		input RESET,
		input [7:0] DI
`ifdef SYS_DRIVEN_CLK4PH
		, input CLK_0
		, input CLK_180
`else // SYS_DRIVEN_CLK2X
		, input CLK2X
		, input CLK
`endif
`ifdef IOR_LARGE_MAPPER
		, input WE_cpcsr
		, input WE_cpdr
//		, input OE_cpcsr
		, input OE_cpdr
		, output [7:0] DO_cpcsr
		, output [7:0] DO_cpdr
`else
		, input WR,RD
		, input [5:0] ADDR
		, output [7:0] DO
`endif
		// external I/O
		, output [7:0] CPI_TXD
		, output CPI_TX
		, input CPI_TX_ACK
		, output CPI_IXS
		, input CPI_IXS_ACK

		, input [7:0] CPI_RXD
		, input CPI_RX
		, output CPI_RX_ACK
		, input CPI_IXR
		, output CPI_IXR_ACK

		// interrupt
		, output CPI_INT
		, input RST_CPI_INT
		);

`ifdef SYS_DRIVEN_CLK4PH
    wire __CLK_180 = 1'b0;
`else // SYS_DRIVEN_CLK2X
    wire __CLK_180 = CLK;
`endif

   wire [7:0] TXD_IN;
   wire [3:0] TX_COUNT;


    wire CPIR; 
    wire CPIE; 
    wire CPIS; 
    wire CPI_TX_CLR; 

    rtavr_fifo fifo (
		.CLR(CPI_TX_CLR)
`ifdef SYS_DRIVEN_CLK4PH
		, .CLK_0(CLK_0)
		, .CLK_180(CLK_180)
`else // SYS_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
		, .CLK(CLK)
`endif
		, .WR(TXD_WR)
		, .RD(CPI_TX_ACK)
		, .OUT(CPI_TXD[7:0])
		, .IN(TXD_IN[7:0])
		, .FULL(TX_FULL)
		, .EMPTY(TX_EMPTY)
		, .COUNT(TX_COUNT[3:0])
		);

`ifdef IOR_LARGE_MAPPER
      assign DO_cpcsr[7:0]  = { CPIR, CPIE, CPIS, CPI_RX, TX_FULL, TX_COUNT[3:2], ~TX_EMPTY};
      assign DO_cpdr[7:0]   = CPI_RXD[7:0];
`else // IOR_LARGE_MAPPER

  wire OE_cpcsr = RD & (ADDR == BASE_ADDR);
  wire WE_cpcsr = WR & (ADDR == BASE_ADDR);
  wire OE_cpdr  = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_cpdr  = WR & (ADDR == (BASE_ADDR + 1));

  wire [7:0] i_DO = OE_cpcsr ? { CPIR, CPIE, CPIS, CPI_RX, TX_FULL, TX_COUNT[3:2], ~TX_EMPTY}
`ifdef OUTPUT_MUX_OR
		: OE_cpdr ? CPI_RXD[7:0] : 0;
`else
		: CPI_RXD[7:0];
`endif

`ifdef OUTPUT_MUX_TRI
  bufif1(DO, i_DO, (OE_cpcsr | OE_cpdr));
`else
  assign DO = i_DO;
`endif
`endif // IOR_LARGE_MAPPER

    reg r_cpir;
    reg r_cpie;
    reg r_cpis;

    assign CPIR  = r_cpir;
    assign CPIE  = r_cpie;
    assign CPIS  = r_cpis;

    assign CPI_TX  = ~TX_EMPTY;
    assign CPI_RX_ACK = OE_cpdr;
    assign TXD_IN = DI;
    assign TXD_WR = WE_cpdr;
    assign CPI_TX_CLR  = WE_cpcsr & DI[0];

    assign CPI_INT   = CPIR  & CPIE;

    assign CPI_IXR_ACK   = RST_CPI_INT;
    assign CPI_IXS       = r_cpis;

  always @(`posedge_CLK_180)
  begin
    if (RESET)
      begin
	r_cpir <= 0;
	r_cpie <= 0;
	r_cpis <= 0;
      end
    else if (~__CLK_180)
      begin
	if (RST_CPI_INT | (WE_cpcsr & DI[7]))   r_cpir <= 0;
	else                                    r_cpir <= CPI_IXR;
	if (WE_cpcsr)                           r_cpie <= DI[6];
	if (CPI_IXS_ACK | (WE_cpcsr & DI[0]))   r_cpis <= 0;
	else if (WE_cpcsr)                      r_cpis <= DI[5];
      end
  end

endmodule

//`define FIFO_DUALPORT

module rtavr_fifo # (
	parameter WIDTH = 8
) (
		input CLR
`ifdef SYS_DRIVEN_CLK4PH
		, input CLK_0
		, input CLK_180
`else // SYS_DRIVEN_CLK2X
		, input CLK2X
		, input CLK
`endif
		, input  [WIDTH-1:0] IN     // DATA IN
		, output [WIDTH-1:0] OUT    // DATA OUT
		, input WR			//
		, input RD			// 
		, output EMPTY			// 
		, output FULL			// 
		, output [3:0] COUNT		// fifo count
		);

`ifdef SYS_DRIVEN_CLK4PH
    wire __CLK_0   = 1'b0;
    wire __CLK_180 = 1'b0;
`else // SYS_DRIVEN_CLK2X
    wire __CLK_0   = ~CLK;
    wire __CLK_180 = CLK;
`endif

`ifdef FIFO_DUALPORT
	reg [3:0] r_pos = 4'h0;
	reg [3:0] w_pos = 4'h0;
	reg [WIDTH-1:0] fifo [0:15];
	reg [WIDTH-1:0] r_out;
	assign OUT = r_out;
	assign EMPTY = ( r_pos == w_pos );
	assign FULL  = ( r_pos == w_pos + 4'h1);
	wire [3:0] fifo_count  = w_pos - r_pos;
	assign COUNT = fifo_count;

	always @(`posedge_CLK_180)
	begin
	  if (CLR)
	    begin
	    end
	  else if (~__CLK_180) 
	    begin
		if (WR) fifo[w_pos] <= IN;
		if (WR &  ~FULL) w_pos <= w_pos + 4'h1;
		if (RD & ~EMPTY) r_pos <= r_pos + 4'h1;
	    end
	end

	always @(`posedge_CLK_0)
	begin
	  if (CLR)
	    begin
		r_out  <= 0;
	    end
	  else if (~__CLK_0) 
	    begin
		if (~EMPTY) r_out <= fifo[r_pos];
	    end
	end

`else // FIFO_SHIFT
	reg [3:0] r_pos = 4'hf;
	reg [WIDTH-1:0] fifo [0:15];
	reg [WIDTH-1:0] r_out;
	assign OUT = r_out;
	assign EMPTY = ( r_pos == 4'hf );
	assign FULL  = ( r_pos == 4'he );
	wire [3:0] fifo_count  = r_pos + 1;
	assign COUNT = fifo_count;

	always @(`posedge_CLK_180)
	begin
	  if (CLR) r_pos <= 4'hf;
	  else if (~__CLK_180)
	    begin
	      if (~WR & RD & ~EMPTY) r_pos <= r_pos - 4'h1;
	      if (WR & ~RD &  ~FULL) r_pos <= r_pos + 4'h1;
	    end
	end
	always @(`posedge_CLK_0)
	begin
	  if (~__CLK_0 & ~EMPTY) r_out <= fifo[r_pos];
	end

  generate 
    genvar i,j;
    for (i=0; i<WIDTH; i = i+1)
      begin : data_in
          for (j=1; j<16; j = j+1)
	  begin : data_in_bit
	     always @(`posedge_CLK_180)
	      begin
		if (WR & ~__CLK_180) fifo [j][i] <= fifo[j-1][i];
	      end
	  end
	     always @(`posedge_CLK_180)
	      begin
		if (WR & ~__CLK_180) fifo [0][i] <= IN[i];
	      end
      end
  endgenerate

`endif

endmodule

`endif // IOR_HAVE_CPI
